@@ -1,11 +1,12 @@
+import json
+import re
 import sys
-
 import logging
 import os
+import textwrap
 from collections import OrderedDict
 import inspect
-from typing import Callable
-import traceback
+from typing import Callable, Iterable, Optional
 
 from celery.canvas import Signature
 from celery.result import AsyncResult
@@ -25,6 +26,7 @@ from firexkit.result import get_tasks_names_from_results, wait_for_any_results,
     RETURN_KEYS_KEY, wait_on_async_results_and_maybe_raise, get_result_logging_name, ChainInterruptedException, \
     ChainRevokedException
 
+
 logger = get_task_logger(__name__)
 
 
@@ -55,7 +57,7 @@ class FireXTask(Task):
     """
     Task object that facilitates passing of arguments and return values from one task to another, to be used in chains
     """
-    DYNAMIC_RETURN = '__dict__'
+    DYNAMIC_RETURN = '__DYNAMIC_RETURN__'
     RETURN_KEYS_KEY = RETURN_KEYS_KEY
 
     def __init__(self):
@@ -94,6 +96,19 @@ class FireXTask(Task):
     def from_plugin(self):
         return self._from_plugin
 
+    @property
+    def task_label(self) -> str:
+        """Returns a label for this task
+
+        Examples:
+            8345379a-e536-4566-b5c9-3d515ec5936a
+            8345379a-e536-4566-b5c9-3d515ec5936a_2 (if it was the second retry)
+            microservices.testsuites_tasks.CreateWorkerConfigFromTestsuites (if there was no request id yet)
+        """
+        label = str(self.request.id) if self.request.id else self.name
+        label += '_%d' % self.request.retries if self.request.retries >= 1 else ''
+        return label
+
     @property
     def request_soft_time_limit(self):
         return self.request.timelimit[1]
@@ -163,21 +178,130 @@ class FireXTask(Task):
         """The body of the task executed by workers."""
         raise NotImplementedError('Tasks must define the run method.')
 
+    @staticmethod
+    def strip_orig_from_name(task_name):
+        return re.sub("(_orig)*$", "", task_name)
+
+    @staticmethod
+    def get_short_name(task_name):
+        # Task name of first task in chain. (I.E. 'task1' in module1.task1|module2.task2)
+        return task_name.split('|')[0].split('.')[-1]
+
+    @property
+    def name_without_orig(self):
+        return self.strip_orig_from_name(self.name)
+
+    @property
+    def short_name(self):
+        return self.get_short_name(self.name)
+
+    @property
+    def short_name_without_orig(self):
+        return self.strip_orig_from_name(self.short_name)
+
+    @property
+    def called_as_orig(self):
+        return True if self.name.endswith('_orig') else False
+
     @abstractmethod
-    def pre_task_run(self):
+    def pre_task_run(self, extra_events: Optional[dict] = None):
         """
         Overrideable method to allow subclasses to do something with the
         BagOfGoodies before returning the results
         """
-        pass
+
+        if extra_events is None:
+            extra_events = {}
+
+        bound_args = self.bound_args
+        default_bound_args = self.default_bound_args
+
+        # Send a custom task-started-info event with the args
+        if not self.request.called_directly:
+            self.send_event('task-started-info',
+                            firex_bound_args=convert_to_serializable(bound_args),
+                            firex_default_bound_args=convert_to_serializable(default_bound_args),
+                            called_as_orig=self.called_as_orig,
+                            long_name=self.name_without_orig,
+                            log_filepath=self.task_logfile,
+                            from_plugin=self.from_plugin,
+                            code_filepath=self.code_filepath,
+                            retries=self.request.retries,
+                            **extra_events)
+
+        # Print the pre-call header
+        self.print_precall_header(bound_args, default_bound_args)
+        self._log_soft_time_limit_override_if_applicable()
+
+    def _log_soft_time_limit_override_if_applicable(self):
+        if not self.request.called_directly:
+            default_soft_time_limit = self.soft_time_limit
+            request_soft_time_limit = self.request_soft_time_limit
+            if default_soft_time_limit != request_soft_time_limit:
+                logger.debug(f'This task default soft_time_limit of '
+                             f'{default_soft_time_limit}{"s" if default_soft_time_limit is not None else ""} '
+                             f'was over-ridden to {request_soft_time_limit}s')
+            else:
+                if default_soft_time_limit is not None:
+                    logger.debug(f'This task soft_time_limit is {default_soft_time_limit}s')
 
     @abstractmethod
-    def post_task_run(self, results):
+    def post_task_run(self, results, extra_events: Optional[dict] = None):
         """
         Overrideable method to allow subclasses to do something with the
         BagOfGoodies after the task has been run
         """
-        pass
+
+        if extra_events is None:
+            extra_events = {}
+
+        # No need to expose the RETURN_KEYS_KEY
+        try:
+            del results[RETURN_KEYS_KEY]
+        except (TypeError, KeyError):
+            pass
+
+        # Print the post-call header
+        self.print_postcall_header(results)
+
+        # Send a custom task-succeeded event with the results
+        if not self.request.called_directly:
+            self.send_event('task-results', firex_result=convert_to_serializable(results), **extra_events)
+
+    def print_precall_header(self, bound_args, default_bound_args):
+        n = 1
+        content = ''
+        args_list = []
+        for postfix, args in zip(['', ' (default)'], [bound_args, default_bound_args]):
+            if args:
+                for k, v in args.items():
+                    args_list.append('  %d. %s: %r%s' % (n, k, v, postfix))
+                    n += 1
+        if args_list:
+            content = 'ARGUMENTS\n' + '\n'.join(args_list)
+
+        task_name = self.name
+        if self.from_plugin:
+            task_name += ' (PLUGIN)'
+
+        logger.debug(banner('STARTED: %s' % task_name, content=content, length=100),
+                     extra={'label': self.task_label, 'span_class': 'task_started'})
+
+    def print_postcall_header(self, result):
+        content = ''
+        results_list = []
+        if result:
+            if isinstance(result, dict):
+                n = 1
+                for k, v in result.items():
+                    results_list.append('  %d. %s: %r' % (n, k, v))
+                    n += 1
+            else:
+                results_list.append('  %r' % result)
+        if results_list:
+            content = 'RETURNS\n' + '\n'.join(results_list)
+        logger.debug(banner('COMPLETED: %s' % self.name, ch='*', content=content, length=100),
+                     extra={'span_class': 'task_completed'})
 
     def __call__(self, *args, **kwargs):
         """
@@ -200,7 +324,17 @@ class FireXTask(Task):
                     logger.debug("The following exception was thrown (and caught) when wait_for_children was "
                                  "implicitly called by this task's base class:\n" + str(e))
             return result
-        except (ChainInterruptedException, ChainRevokedException) as e:
+        except Exception as e:
+            self.handle_exception(e)
+        finally:
+            try:
+                if self._lagging_children_strategy is not PendingChildStrategy.Continue:
+                    self.revoke_pending_children()
+            finally:
+                self.remove_task_logfile_handler()
+
+    def handle_exception(self, e, logging_extra=None, raise_exception=True):
+        if isinstance(e, ChainInterruptedException) or isinstance(e, ChainRevokedException):
             try:
                 exception_cause_uuid = e.task_id
             except AttributeError:
@@ -208,19 +342,13 @@ class FireXTask(Task):
             else:
                 if exception_cause_uuid:
                     self.send_event('task-exception-cause', exception_cause_uuid=exception_cause_uuid)
-            logger.debug(e, exc_info=True)
-            logger.error(e)
-            raise
-        except Exception as e:
-            logger.debug(e, exc_info=True)
-            logger.error(e)
-            raise
-        finally:
-            try:
-                if self._lagging_children_strategy is not PendingChildStrategy.Continue:
-                    self.revoke_pending_children()
-            finally:
-                self.remove_task_logfile_handler()
+        mssg = f'{type(e).__name__}'
+        exception_string = str(e)
+        if exception_string:
+            mssg += f': {exception_string}'
+        logger.error(mssg, exc_info=e, extra=logging_extra)
+        if raise_exception:
+            raise e
 
     def _process_arguments_and_run(self, *args, **kwargs):
         # Organise the input args by creating a BagOfGoodies
@@ -260,9 +388,9 @@ class FireXTask(Task):
 
         if not self.request.called_directly:
             if self.request.retries == self.max_retries:
-                logger.error('Failed all %d retry attempts' % self.max_retries)
+                logger.error(f'{self.short_name} failed all {self.max_retries} retry attempts')
             else:
-                logger.warning('Failed and retrying %d/%d' % (self.request.retries+1, self.max_retries))
+                logger.warning(f'{self.short_name} failed and retrying {self.request.retries+1}/{self.max_retries}')
         super(FireXTask, self).retry(*args, **kwargs)
 
     @property
@@ -382,9 +510,17 @@ class FireXTask(Task):
                 [self._update_child_state(child_result, self._UNBLOCKED) for child_result in child_results]
 
     def enqueue_child(self, chain: Signature, add_to_enqueued_children: bool = True, block: bool = False,
+                      raise_exception_on_failure: bool = None,
                       apply_async_epilogue: Callable[[AsyncResult], None] = None, apply_async_options=None,
                       **kwargs) -> AsyncResult:
         """Schedule a child task to run"""
+
+        if raise_exception_on_failure is not None:
+            if not block:
+                raise ValueError('Cannot control exceptions on child failure if we don\'t block')
+            # Only set it if not None, otherwise we want to leave the downstream default
+            kwargs['raise_exception_on_failure'] = raise_exception_on_failure
+
         if apply_async_options is None:
             apply_async_options = dict()
 
@@ -402,7 +538,8 @@ class FireXTask(Task):
         if block:
             try:
                 wait_on_async_results_and_maybe_raise(results=child_result,
-                                                      caller_task=self, **kwargs)
+                                                      caller_task=self,
+                                                      **kwargs)
             finally:
                 if add_to_enqueued_children:
                     self._update_child_state(child_result, self._UNBLOCKED)
@@ -566,3 +703,42 @@ def get_attr_unwrapped(fun: callable, attr_name, *default_value):
     if default_value:
         return default_value[0]
     raise AttributeError(attr_name)
+
+
+def is_jsonable(obj) -> bool:
+    """Returns :const:`True` if the `obj` can be serialized via Json,
+    otherwise returns :const:`False`
+    """
+    try:
+        json.dumps(obj)
+    except TypeError:
+        return False
+    else:
+        return True
+
+
+def convert_to_serializable(obj, depth=0):
+    if is_jsonable(obj):
+        return obj
+
+    # recursive reference guard.
+    if depth < 10:
+        # Full object isn't jsonable, but some contents might be. Try walking the structure to get jsonable parts.
+        if isinstance(obj, dict):
+            return {k: convert_to_serializable(v, depth+1) for k, v in obj.items()}
+
+        # Note that it's important this DOES NOT catch strings, and it won't since strings are jsonable.
+        if isinstance(obj, Iterable):
+            return [convert_to_serializable(e, depth+1) for e in obj]
+
+    # Either input isn't walkable (i.e. dict or iterable), or we're too deep in the structure to keep walking.
+    return repr(obj)
+
+
+def banner(text, ch='=', length=78, content=''):
+    if content:
+        content = '\n'.join(['\n'.join(textwrap.wrap(line, width=length)) for line in content.split('\n')])
+        content += '\n'
+    spaced_text = '\n'.join(
+        ['\n'.join(textwrap.wrap(line, width=length, drop_whitespace=False)) for line in text.split('\n')])
+    return '\n' + ch * length + '\n' + spaced_text.center(length, ch) + '\n' + content + ch * length + '\n'
